package com.itheima.util;

import java.util.Arrays;
import java.util.Iterator;
import java.util.NoSuchElementException;

/**
 * 此类实现了对队列数据结构的模仿；此类还实现了Iterable接口，为遍历队列元素提供了方便。内部封装的是数组
 * 
 * @author mrng
 * 
 * @param <E>
 */
@SuppressWarnings("unchecked")
public class LoopList<E> implements Iterable<E> {
	/**
	 * 队列的默认长度
	 */
	private static final int DEFAULT_MAX_SIZE = 5;

	/**
	 * 队列的最大长度
	 */
	private int maxSize;

	/**
	 * 存储队列数据的内部容器
	 */
	private Object[] list;

	/**
	 * 队列指针
	 */
	private int pos;

	/**
	 * 创建一个长度为10的队列
	 */
	public LoopList() {
		this.maxSize = DEFAULT_MAX_SIZE;
		list = new Object[this.maxSize];
	}

	/**
	 * 通过maxSiz指定此队列的最大长度
	 * 
	 * @param maxSize
	 *            队列的最大长度
	 */
	public LoopList(int maxSize) {
		this.maxSize = maxSize;
		list = new Object[this.maxSize];
	}

	/**
	 * 此方法可以将队列整体向前移动一位，第一个将被挤出队列
	 */
	private void leftMove() {
		for (int i = 0; i < list.length - 1; i++) {
			list[i] = list[i + 1];
		}
	}

	/**
	 * 添加元素，如果队列的长度已达到最大长度，则将第一个挤出队列并返回，再将新的元素添加进队列
	 * 
	 * @param e
	 *            要添加的新元素
	 * @return 如果有被挤出的元素则将其返回，否则返回null
	 */
	public E add(E e) {
		// 如果队列的长度已达到最大长度，则将第一个挤出队列并返回，再将新的元素添加进队列
		if (pos == list.length) {
			E last = (E) list[0];
			leftMove();
			list[list.length - 1] = e;
			return last;
		}
		list[pos++] = e;
		return null;
	}

	/**
	 * 移除此队列中的所有元素
	 */
	public void clear() {
		for (int i = 0; i < pos; i++) {
			list[i] = null;
		}
		pos = 0;
	}

	/**
	 * 如果此队列包含指定元素，则返回 true
	 * 
	 * @param obj
	 *            要测试在此队列中是否存在的元素
	 * @return 如果此队列包含指定元素，则返回 true
	 */
	public boolean contains(Object obj) {
		for (int i = 0; i < pos; i++) {
			// if (((E)list[i]).equals((E)obj))
			if (list[i].equals(obj))
				return true;
		}
		return false;
	}

	/**
	 * 获取但不移除此队列的第一个元素；如果此队列为空，则返回null
	 * 
	 * @return 队列的第一个元素，如果此列表为空，则返回null
	 */
	public E getFirst() {
		return (E) (pos > 0 ? list[0] : null);
	}

	/**
	 * 获取但不移除此队列的最后一个元素；如果此队列为空，则返回null
	 * 
	 * @return 队列的最后一个元素，如果此列表为空，则返回null
	 */
	public E getLast() {
		return (E) (pos > 0 ? list[pos - 1] : null);
	}

	/**
	 * 移除并返回此队列的第一个元素；如果此队列为空，则返回null
	 * 
	 * @return 队列的第一个元素，如果此列表为空，则返回null
	 */
	public E removeFirst() {
		if (pos > 0) {
			E first = (E) list[0];
			leftMove();
			pos--;
			return first;
		}
		return null;
	}

	/**
	 * 移除并返回此队列的最后一个元素；如果此队列为空，则返回null
	 * 
	 * @return 队列的最后一个元素，如果此列表为空，则返回null
	 */
	public E removeLast() {
		if (pos > 0) {
			E last = (E) list[pos - 1];
			list[pos - 1] = null;
			pos--;
			return last;
		}
		return null;
	}

	/**
	 * 获取队列中指定位置上的元素
	 * 
	 * @param index
	 *            - 要返回的元素的索引
	 * @return 队列中指定位置的元素
	 * 
	 * @exception IndexOutOfBoundsException
	 *                如果索引超出范围 (index < 0 || index >= size())
	 */
	public E getChildAt(int index) {
		if (index < 0 || index >= pos)
			throw new IndexOutOfBoundsException();

		return (E) list[index];
	}

	/**
	 * 将此队列中指定位置的元素替换为指定的元素
	 * 
	 * @param index
	 *            要替换的元素的索引
	 * @param element
	 *            要在指定位置存储的元素
	 * @return 返回以前在指定位置上的元素
	 * 
	 * @exception IndexOutOfBoundsException
	 *                如果索引超出范围 (index < 0 || index >= size())
	 */
	public E set(int index, E element) {
		if (index < 0 || index >= pos)
			throw new IndexOutOfBoundsException();

		E old = (E) list[index];
		list[index] = element;
		return old;
	}

	/**
	 * 按队列中的顺序返回一个包含所有元素的数组
	 * 
	 * @return 按队列中的顺序返回一个包含所有元素的数组
	 */
	public Object[] toArray() {
		return Arrays.copyOf(list, list.length);
	}

	/**
	 * 按队列中的顺序返回一个包含所有元素的数组 返回数组的运行时类型为指定数组的类型。 如果指定数组能容纳队列，则在其中返回该队列。
	 * 否则，分配具有指定数组的运行时类型和此队列大小的新数组。
	 * 
	 * @param a
	 *            要在其中存储队列元素的数组（如果它足够大）；否则，分配具有指定数组的运行时类型和此队列大小的新数组
	 * @return 按队列中的顺序返回一个包含所有元素的数组
	 * 
	 * @exception ArrayStoreException
	 *                如果指定数组的运行时类型不是此队列中每个元素的运行时类型的超类型
	 * @exception NullPointerException
	 *                如果指定的数组为 null
	 */
	public <T> T[] toArray(T[] a) {
		// 如果a的长度小于队列的长度，就创建长度和队列一致的新的T类型的数组
		if (a.length < pos)
			a = (T[]) java.lang.reflect.Array.newInstance(a.getClass()
					.getComponentType(), pos);

		// 将队列中的元素拷贝到a中
		for (int i = 0; i < pos; i++)
			a[i] = (T) list[i];

		// 如果a的长度大于队列的长度，则将a中队列元素的最后一个元素的后面的元素置为null，表示队列的结尾
		if (a.length > pos)
			a[pos] = null;

		return a;
	}

	/**
	 * 获取队列的长度
	 * 
	 * @return 队列的长度
	 */
	public int getSize() {
		return pos;
	}

	/**
	 * 返回次队列的最大长度
	 * 
	 * @return 此队列的最大长度
	 */
	public int getMaxSize() {
		return this.maxSize;
	}

	/**
	 * 如果此队列中没有元素，则返回true
	 * 
	 * @return 如果此队列中没有元素，则返回true
	 */
	public boolean isEmpty() {
		return pos == 0 ? true : false;
	}

	/**
	 * 返回此队列的字符串表现形式
	 */
	public String toString() {
		StringBuilder sb = new StringBuilder("[");
		for (Object o : list) {
			sb.append(o + ",");
		}
		sb.append("]");

		return sb.toString();
	}

	/**
	 * 返回此队列的迭代器
	 */
	public Iterator<E> iterator() {
		return new MyIterator();
	}

	/**
	 * 此类为队列提供了迭代功能
	 * 
	 * @author mrng
	 * 
	 */
	private class MyIterator implements Iterator<E> {
		/**
		 * 此队列迭代器的指针
		 */
		private int itpos = 0;

		/**
		 * 表示是否可以执行remove操作
		 */
		private boolean lastRet = false;

		/**
		 * 如果队列迭代器中有元素，则返回 true
		 */
		public boolean hasNext() {
			if (itpos < pos)
				return true;
			return false;
		}

		/**
		 * 返回次队列迭代器中的下一个元素
		 * 
		 * @exception NoSuchElementException
		 *                如果没有元素可迭代则抛出此异常
		 */
		public E next() {
			// 如果队列中没有元素或迭代器指针已指到队列后面，则抛出次异常
			if (itpos == pos)
				throw new NoSuchElementException();

			// 如果next可以返回元素，则将remove操作解锁
			lastRet = true;
			return (E) list[itpos++];
		}

		/**
		 * 从迭代器指向的 队列 中移除迭代器返回的最后一个元素。 每次调用 next 后只能调用一次此方法。
		 * 如果进行迭代时用调用此方法之外的其他方式修改了该迭代器所指向的 队列，则迭代器的行为是不确定的。
		 * 
		 * @exception IllegalStateException
		 *                - 如果尚未调用 next 方法，或者在上一次调用 next 方法之后已经调用了 remove 方法。
		 */
		public void remove() {
			// 如果不可以执行remove操作，则抛出此异常
			if (!lastRet)
				throw new IllegalStateException();

			// 删除元素后，重置list队列
			for (int i = itpos - 1; i < pos - 1; i++)
				list[i] = list[i + 1];

			// 删除元素后，将remove操作冻结，这样就避免了重复执行remove操作
			lastRet = false;
			// 删除元素后，由于list队列重置(被删除元素后面的所有元素前移了一位)，所以将itpos -1
			itpos--;
			// 对队列进行重置后，将最后一个重复元素置空
			list[--pos] = null;
		}
	}
}
